#include "game/Enemy.hpp"

#include <cmath>
#include <ctime>
#include <SFML/Graphics/RenderWindow.hpp>
#include <util/Math.hpp>

#include "scenes/GameScene.hpp"
#include "game/Player.hpp"
#include "game/Projectile.hpp"
#include "game/World.hpp"

namespace
{
	float getSignOf( float f )
	{
		return f / std::abs( f );
	}
}

namespace game
{
	Enemy::Enemy( Game& theGame, GameScene& theScene, World& theWorld, sf::Vector2f pos )
	   : Object::Object( theGame, theScene, theWorld ),
	     coll( 4, 4, 8, 8 ),
	     vel( 0, 0 ),
	     shoot( 1 ),
	     timeAlive( 0 )
	{
		shape.setPointCount( 4 );
		shape.setPoint( 0, sf::Vector2f( 8, 18 ) );
		shape.setPoint( 1, sf::Vector2f( 16, 0 ) );
		shape.setPoint( 2, sf::Vector2f( 8, 4 ) );
		shape.setPoint( 3, sf::Vector2f( 0, 0 ) );
		shape.setOrigin( 8, 8 );
		shape.setFillColor( sf::Color( 255, 32, 0 ) );
		shape.setOutlineColor( sf::Color( 128, 16, 0, 128 ) );
		shape.setOutlineThickness( 1.5f );
		shape.setPosition( pos );
		
		if ( rand() % 25 == 0 )
		{
			ai = AiType::Still;
		}
		else
		{
			ai = static_cast< AiType >( rand() % ( AiType::Count - 1 ) );
		}
		canShoot = ( ( ai == AiType::Still ) ? true : ( ( rand() % 4 ) == 0 ) );
		orbitDir = ( ( rand() % 2 ) == 0 ) ? -1 : 1;
		accuracy = 0.5f + ( rand() % 20 ) / 40.f;
		swerviness = 1 + ( rand() % 4 );
		speediness = 0.75 + ( ( rand() % 20 ) / 27.5f );
		bulletSpeed = ( rand() % 25 ) / 10.f;
		reloadTime = 30 + ( rand() % 45 );
	}
	
	void Enemy::update()
	{
		++timeAlive;
		if ( timeAlive >= 50 * 5 )
		{
			sf::View view = scene.getPlayer().getView();
			sf::FloatRect viewRect( view.getCenter().x - ( view.getSize().x / 2 ), view.getCenter().y - ( view.getSize().y / 2 ), view.getSize().x, view.getSize().y );
			
			if ( !viewRect.contains( shape.getPosition() ) )
			{
				world.removeObject( this );
			}
		}
		
		Player& player = scene.getPlayer();
		sf::Vector2f playerPos = player.getPosition();
		
		for ( auto it = world.getObjects().begin(); it != world.getObjects().end(); ++it )
		{
			Object* obj = it->get();
			if ( Projectile* proj = dynamic_cast< Projectile* >( obj ) )
			{
				if ( proj->isPlayer() and proj->getCollisionRect().intersects( getCollisionRect() ) and proj->getLife() > 0 )
				{
					++scene.points;
					world.removeObject( this );
					proj->damage();
					player.playEnemyHurt();
					return;
				}
			}
		}
		
		sf::Vector2f pos = shape.getPosition();
		float dir = util::radiansToDegrees( std::atan2( playerPos.x - pos.x, playerPos.y - pos.y ) );
		shape.setRotation( -dir );
		
		if ( ai != AiType::Still )
		{
			const float SPEED = 0.5f * ( 0.75 + scene.getDifficulty() * 0.5f ) * speediness;
			float factor = 1;
			
			float walkDir = dir;// + 180;
			if ( ai == AiType::Swerve )
			{
				constexpr int RANGE = 135;
				float swerveDir = walkDir + ( ( timeAlive % ( RANGE * swerviness ) ) / static_cast< float >( swerviness ) ) - ( RANGE / 2 );
				
				vel.y += std::cos( util::degreesToRadians( swerveDir ) ) * SPEED / 2;
				vel.x += std::sin( util::degreesToRadians( swerveDir ) ) * SPEED / 2;
				factor = 0.5f;
				
				shape.setRotation( -( ( swerveDir + walkDir ) / 2 ) );
			}
			else if ( ai == AiType::Orbit )
			{
				walkDir += 67 * orbitDir;
			}
			
			// Don't ask why these are swapped. I have no clue
			vel.y += std::cos( util::degreesToRadians( walkDir ) ) * SPEED * factor;
			vel.x += std::sin( util::degreesToRadians( walkDir ) ) * SPEED * factor;
			
			const float MAX_VELOCITY = 3.f * ( 0.75 + scene.getDifficulty() * 0.2f );
			if ( std::abs( vel.x ) > MAX_VELOCITY ) vel.x = MAX_VELOCITY * getSignOf( vel.x );
			if ( std::abs( vel.y ) > MAX_VELOCITY ) vel.y = MAX_VELOCITY * getSignOf( vel.y );
			
			// Friction
			constexpr float FRICTION = 0.1f * ( 0.75 + scene.getDifficulty() * 0.5f );
			if ( vel.x != 0 ) vel.x -= FRICTION * getSignOf( vel.x );
			if ( vel.y != 0 ) vel.y -= FRICTION * getSignOf( vel.y );
			
			if ( std::abs( vel.x ) < FRICTION + 0.01f ) vel.x = 0.f;
			if ( std::abs( vel.y ) < FRICTION + 0.01f ) vel.y = 0.f;
			
			shape.move( vel );
			
			// World limits
			sf::Vector2f pos = shape.getPosition();
			sf::FloatRect bounds = world.getBounds();
			
			if ( pos.x < bounds.left ) pos.x = bounds.left;
			if ( pos.y < bounds.top ) pos.y = bounds.top;
			if ( pos.x > bounds.left + bounds.width ) pos.x = bounds.left + bounds.width;
			if ( pos.y > bounds.top + bounds.height ) pos.y = bounds.top + bounds.height;
			
			shape.setPosition( pos );
		}
		
		if ( canShoot and --shoot <= 0 )
		{
			if ( ai == AiType::Still )
			{
				shape.setRotation( dir );
			}
			
			float shootDir = dir;
			shootDir += ( ( rand() % 360 ) - 180 ) / ( 1.f - accuracy );
			
			Projectile* proj = new Projectile( game, scene, world, shape.getPosition(), shootDir, 5.f + bulletSpeed, false, 1 );
			world.addObject( proj );
			
			shoot = reloadTime;
		}
	}

	void Enemy::render( sf::RenderWindow& window )
	{
		window.draw( shape );
	}
	
	sf::FloatRect Enemy::getCollisionRect() const
	{
		sf::FloatRect ret = coll;
		ret.left += shape.getPosition().x - shape.getOrigin().x;
		ret.top += shape.getPosition().y - shape.getOrigin().y;
		
		return ret;
	}
}
