#include "game/Player.hpp"

#include <cmath>
#include <SFML/Graphics/RenderWindow.hpp>
#include <SFML/Window/Keyboard.hpp>
#include <SFML/Window/Mouse.hpp>
#include <util/Math.hpp>

#include "Game.hpp"
#include "game/Enemy.hpp"
#include "game/Projectile.hpp"
#include "game/World.hpp"
#include "scenes/GameScene.hpp"

namespace
{
	float getSignOf( float f )
	{
		return f / std::abs( f );
	}
}

namespace game
{
	Player::Player( Game& theGame, GameScene& theScene, World& theWorld )
	   : Object::Object( theGame, theScene, theWorld ),
	     damage( 0 ),
	     invuln( 0 ),
	     coll( 4, 4, 8, 8 ),
	     vel( 0, 0 ),
	     shoot( 1 )
	{
		shape.setPointCount( 5 );
		shape.setPoint( 0, sf::Vector2f( 8, 18 ) );
		shape.setPoint( 1, sf::Vector2f( 10, 15 ) );
		shape.setPoint( 2, sf::Vector2f( 16, 0 ) );
		shape.setPoint( 3, sf::Vector2f( 0, 0 ) );
		shape.setPoint( 4, sf::Vector2f( 6, 15 ) );
		shape.setOrigin( 8, 8 );
		shape.setFillColor( sf::Color( 0, 128, 255 ) );
		shape.setOutlineColor( sf::Color( 0, 64, 128, 128 ) );
		shape.setOutlineThickness( 1.5f );
		shape.setPosition( 320, 240 );
	}

	void Player::update()
	{
		updateView();
		updateDirection();
		updatePosition();
		
		checkHurt();
		checkShooting();
	}

	void Player::render( sf::RenderWindow& window )
	{
		if ( invuln == 0 or invuln % 10 < 5 )
		{
			window.draw( shape );
		}
	}
	
	int Player::getMaxHealth() const
	{
		return 1 + ( ( skills[ Skill::Health ] - 1 ) * 2 );
	}
	
	int Player::getDamage() const
	{
		return damage;
	}
	
	sf::Vector2f Player::getPosition() const
	{
		return shape.getPosition();
	}
	
	sf::FloatRect Player::getCollisionRect() const
	{
		sf::FloatRect ret = coll;
		ret.left += shape.getPosition().x - shape.getOrigin().x;
		ret.top += shape.getPosition().y - shape.getOrigin().y;
		
		return ret;
	}
	
	sf::View Player::getView() const
	{
		return view;
	}
	
	void Player::playEnemyHurt()
	{
		enemyHurtSound.setBuffer( * game.getSoundBuffer( "res/sound/enemyHurt.wav" ) );
		enemyHurtSound.setVolume( 25 );
		enemyHurtSound.play();
	}
	
	void Player::updateView()
	{
		sf::Vector2u windowSize = game.getWindow().getSize();
		
		sf::FloatRect bounds = world.getBounds();
		sf::FloatRect camBounds = bounds;
		camBounds.left += windowSize.x / 2;
		camBounds.top += windowSize.y / 2;
		camBounds.width -= windowSize.x;
		camBounds.height -= windowSize.y;
		
		sf::Vector2f center = shape.getPosition();
		if ( center.x < camBounds.left ) center.x = camBounds.left;
		if ( center.y < camBounds.top ) center.y = camBounds.top;
		if ( center.x > camBounds.left + camBounds.width ) center.x = camBounds.left + camBounds.width;
		if ( center.y > camBounds.top + camBounds.height ) center.y = camBounds.top + camBounds.height;
		
		view = sf::View( sf::FloatRect( 0, 0, windowSize.x, windowSize.y ) );
		view.setCenter( center );
	}
	
	void Player::updateDirection()
	{
		sf::Vector2i mouseRaw = sf::Mouse::getPosition( game.getWindow() );
		sf::Vector2f mouseWorld = game.getWindow().mapPixelToCoords( mouseRaw, view );
		sf::Vector2f mouseLocal( mouseWorld.x - shape.getPosition().x, mouseWorld.y - shape.getPosition().y );
		
		float dir = -std::atan2( mouseLocal.x, mouseLocal.y );
		shape.setRotation( util::radiansToDegrees( dir ) );
	}
	
	void Player::updatePosition()
	{
		// Basic movement
		const float SPEED = 0.5f * ( 1 + skills[ Skill::Speed ] * 0.5f );
		if ( sf::Keyboard::isKeyPressed( sf::Keyboard::W ) or sf::Keyboard::isKeyPressed( sf::Keyboard::Up ) )
		{
			vel.y -= SPEED;
		}
		if ( sf::Keyboard::isKeyPressed( sf::Keyboard::S ) or sf::Keyboard::isKeyPressed( sf::Keyboard::Down ) )
		{
			vel.y += SPEED;
		}
		if ( sf::Keyboard::isKeyPressed( sf::Keyboard::A ) or sf::Keyboard::isKeyPressed( sf::Keyboard::Left ) )
		{
			vel.x -= SPEED;
		}
		if ( sf::Keyboard::isKeyPressed( sf::Keyboard::D ) or sf::Keyboard::isKeyPressed( sf::Keyboard::Right ) )
		{
			vel.x += SPEED;
		}
		
		// Maximum velocity
		const float MAX_VELOCITY = 3.f * ( 1 + skills[ Skill::Speed ] * 0.2f );
		if ( std::abs( vel.x ) > MAX_VELOCITY ) vel.x = MAX_VELOCITY * getSignOf( vel.x );
		if ( std::abs( vel.y ) > MAX_VELOCITY ) vel.y = MAX_VELOCITY * getSignOf( vel.y );
		
		// Friction
		constexpr float FRICTION = 0.1f * ( 1 + skills[ Skill::Speed ] * 0.5f );
		if ( vel.x != 0 ) vel.x -= FRICTION * getSignOf( vel.x );
		if ( vel.y != 0 ) vel.y -= FRICTION * getSignOf( vel.y );
		
		if ( std::abs( vel.x ) < FRICTION + 0.01f ) vel.x = 0.f;
		if ( std::abs( vel.y ) < FRICTION + 0.01f ) vel.y = 0.f;
		
		shape.move( vel );
		
		// World limits
		sf::Vector2f pos = shape.getPosition();
		sf::FloatRect bounds = world.getBounds();
		
		if ( pos.x < bounds.left ) pos.x = bounds.left;
		if ( pos.y < bounds.top ) pos.y = bounds.top;
		if ( pos.x > bounds.left + bounds.width ) pos.x = bounds.left + bounds.width;
		if ( pos.y > bounds.top + bounds.height ) pos.y = bounds.top + bounds.height;
		
		shape.setPosition( pos );
	}
	
	void Player::checkHurt()
	{
		hurtSound.setBuffer( * game.getSoundBuffer( "res/sound/playerHurt.wav" ) );
		hurtSound.setVolume( 100 );
		
		if ( invuln > 0 )
		{
			--invuln;
			return;
		}
		
		sf::FloatRect collision = getCollisionRect();
		
		for ( auto it = world.getObjects().begin(); it != world.getObjects().end(); ++it )
		{
			Object* obj = it->get();
			if ( Projectile* proj = dynamic_cast< Projectile* >( obj ) )
			{
				if ( proj->getCollisionRect().intersects( collision ) and !proj->isPlayer() and invuln <= 0 )
				{
					proj->damage();
					++damage;
					invuln = 25 - ( 5 * scene.getDifficulty() );
					hurtSound.play();
					break;
				}
			}
			else if ( Enemy* enemy = dynamic_cast< Enemy* >( obj ) )
			{
				if ( enemy->getCollisionRect().intersects( collision ) and invuln <= 0 )
				{
					world.removeObject( enemy );
					++damage;
					invuln = 25 - ( 5 * scene.getDifficulty() );
					hurtSound.play();
					break;
				}
			}
		}
	}
	
	void Player::checkShooting()
	{
		// Above button check so it reloads even while not shooting
		if ( shoot > 0 and --shoot > 0 ) // Only decrement if above zero
		{
			return;
		}
		
		if ( !sf::Mouse::isButtonPressed( sf::Mouse::Left ) && !sf::Keyboard::isKeyPressed( sf::Keyboard::Space ) )
		{
			return;
		}
		
		const int SHOOT_TIME = 23 - ( skills[ Skill::Reload ] * 3 );
		
		shoot = SHOOT_TIME;
		world.addObject( new Projectile( game, scene, world, shape.getPosition(), shape.getRotation(), 6.5f, true, skills[ Skill::Pierce ] ) );
	}
}
