#include "scenes/GameScene.hpp"

#include <algorithm>
#include <SFML/Graphics/Font.hpp>
#include <SFML/Graphics/RectangleShape.hpp>
#include <SFML/Graphics/RenderWindow.hpp>
#include <SFML/Graphics/Text.hpp>
#include <util/String.hpp>

#include "Game.hpp"
#include "game/Enemy.hpp"
#include "game/Player.hpp"

GameScene::GameScene( Game& theGame )
   : Scene::Scene( theGame )
{
}

void GameScene::initialize( SceneChangeEvent& event )
{
	world.reset( new game::World() );
	
	player = new game::Player( game, ( * this ), ( * world ) );
	player->skills = event.skills;
	world->addObject( player );
	
	points = 0.f;
	difficulty = event.difficulty;
	spawnTimer = 0;
	finishTimer = ( 10 + 1 ) * 50;
	
	std::srand( 0 );
}

void GameScene::terminate()
{
	world.reset();
	player = NULL;
}

void GameScene::update()
{
	world->update();
	
	if ( ++spawnTimer >= 25 )
	{
		sf::FloatRect bounds = world->getBounds();
		sf::FloatRect view( player->getPosition().x - 320 - 8, player->getPosition().y - 240 - 8, 640 + 16, 480 + 16 );
		
		for ( int i = 0; i < 7 + ( 2 * difficulty ); ++i )
		{
			float x = rand() % static_cast< int >( bounds.width );
			float y = rand() % static_cast< int >( bounds.height );
			x += bounds.left;
			y += bounds.top;
			
			if ( view.contains( x, y ) )
			{
				continue;
			}
			
			game::Enemy* enemy = new game::Enemy( game, ( * this ), ( * world ), sf::Vector2f( x, y ) );
			world->addObject( enemy );
			
			spawnTimer = 0;
		}
	}
	
	if ( player->getDamage() == player->getMaxHealth() )
	{
		game.changeScenes( "Game Over" );
	}
	
	if ( finishTimer >= 0 )
	{
		--finishTimer;
	}
	else
	{
		SceneChangeEvent event;
		event.difficulty = difficulty;
		event.score = points;
		
		game.changeScenes( "Submit Score", event );
	}
}

void GameScene::update( const sf::Event& event )
{
	world->update( event );
}

void GameScene::render( sf::RenderWindow& window )
{
	window.clear( sf::Color::White );
	
	window.setView( player->getView() );
	{
		sf::RectangleShape shape;
		shape.setFillColor( sf::Color( 0, 0, 0, 0 ) );
		shape.setOutlineColor( sf::Color::Black );
		shape.setOutlineThickness( 3.f );
		shape.setPosition( world->getBounds().left + 3, world->getBounds().top + 3 );
		shape.setSize( sf::Vector2f( world->getBounds().width - 6, world->getBounds().height - 6 ) );
		window.draw( shape );
		
		world->render( window );
	}
	
	window.setView( window.getDefaultView() );
	{
		sf::Font& font = ( * game.getFont( "res/dejavu/ttf/DejaVuSans.ttf" ) );
		
		sf::Text score;
		score.setFont( font );
		score.setCharacterSize( 15.f );
		score.setString( "Score: " + util::toString( points ) );
		score.setPosition( 5, 5 );
		score.setColor( sf::Color::Black );
		window.draw( score );
		
		sf::RectangleShape shape;
		shape.setFillColor( sf::Color::Red );
		shape.setOutlineColor( sf::Color::Black );
		shape.setOutlineThickness( 3.f );
		shape.setPosition( 8, 480 - 8 - 16 );
		shape.setSize( sf::Vector2f( player->getMaxHealth() * 16, 16 ) );
		window.draw( shape );
		
		shape.setFillColor( sf::Color::Green );
		shape.setOutlineThickness( 0.f );
		shape.setSize( sf::Vector2f( ( player->getMaxHealth() - player->getDamage() ) * 16, 16 ) );
		window.draw( shape );
		
		sf::Text time = score;
		time.setCharacterSize( 30 );
		time.setString( util::toString( std::min( 10, ( finishTimer / 50 ) + 1 ) ) );
		time.setPosition( 640 - time.getLocalBounds().width - 10, 5 );
		window.draw( time );
	}
}

int GameScene::getDifficulty() const
{
	return difficulty;
}

game::Player& GameScene::getPlayer()
{
	return ( * player );
}

const game::Player& GameScene::getPlayer() const
{
	return ( * player );
}
